package fct.ample.REVMLInterpretor.FMPUtils;


import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.SortedMap;
import java.util.TreeMap;

import org.jdom.Attribute;
import org.jdom.DataConversionException;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.Namespace;
import org.jdom.input.SAXBuilder;


//import Copy;
//import net.ample.tracing.framework.core.exceptions.InvalidXMLFileException;


/**
 * This class is responsible for importing the configs of Feature models written in FMP format.
 * @author Andre Sousa
 */
public class ImportFmpConfigs {
	static final long serialVersionUID = 4;

	private Document document;

	/**
	 * The constructor for ImportFmpFile.
	 * @param fmpFile A String with the path for the FMP file to be imported.  
	 */
	public ImportFmpConfigs(String fmpFile) throws InvalidXMLFileException {
		try {
			File f = new File(fmpFile);
			this.document = this.loadDocument(f);
		}
		catch (IOException e) {
			e.printStackTrace();
		}
	}

	/**
	 * This method loads the file into a JDOM document representing the FMP model.
	 * @param file The file to be read.
	 * @return A JDOM document representing the XML read from the file.
	 * @throws IOException
	 */
	private Document loadDocument(File file) throws IOException, InvalidXMLFileException {
		SAXBuilder builder = new SAXBuilder();
		Document doc = new Document();
		try	{
			doc = builder.build(file);
		}
		catch(JDOMException e) {
			String message = file.getName() + " does not contains well formed XML!\n" + e.getMessage();
			throw new InvalidXMLFileException(message);
		}
		return doc;
	}

	/**
	 * This function returns a Namespace object with the given String prefix.
	 * @param ns The prefix for the namespace.
	 * @return The Namespace represented by the prefix.
	 */
	private Namespace getFmpNamespace(String ns) {
		return document.getRootElement().getNamespace(ns);
	}

	/**
	 * This function returns an FeatureAttributes object containing the attributes (name, min, max) of a given feature.
	 * @param propertiesId The properties id of the feature for which we want to get the the attributes.
	 * @return A FeatureAttributes of the required feature, or null if there are none.
	 */
	private FeatureProperties getFmpFeatureProperties(String propertiesId) {
		try {
			Attribute name = new Attribute("name","null");
			Attribute min = new Attribute("min","1");
			Attribute max = new Attribute("max","1");
			Element metamodel = document.getRootElement().getChild("metaModel");
			//browse through all the <configuration> elements that are inside <metaModel> element.
			for(Iterator<Element> configsIt = metamodel.getDescendants(new ElementNameFilter("configurations")); configsIt.hasNext();) {
				Element current = configsIt.next();
				//it is the node we were looking for.
				if(propertiesId.equals(current.getAttributeValue("id",getFmpNamespace("xmi")))) {
					List<Element> children = current.getChildren("children");
					//go through all the children and get the values we need from each.
					for(Iterator<Element> childrenIt = children.iterator(); childrenIt.hasNext();) {
						Element child = childrenIt.next();
						if("name".equalsIgnoreCase(child.getAttributeValue("name")))
							name = child.getChild("typedValue").getAttribute("stringValue");
						else if("min".equalsIgnoreCase(child.getAttributeValue("name")))
							min = child.getChild("typedValue").getAttribute("integerValue");
						else if("max".equalsIgnoreCase(child.getAttributeValue("name")))
							max = child.getChild("typedValue").getAttribute("integerValue");
					}
					return new FeatureProperties(name.getValue(),min.getIntValue(),max.getIntValue());
				}
			}
			return null;
		}
		catch (DataConversionException e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * This method returns the state associated with a a given feature.
	 * @param config the configuration {@link Element} that is to be parsed.
	 * @param id the ID of the desired feature.
	 * @return A {@link String} with the state. <i>selected</i> if the feature is selected, <i>eliminated</i> if the feature is eliminated, <i>undefined</i> if no state was found or null if no feature was found.
	 */
	private String getFmpFeatureStateAux(Element config, String id) {
		for(Iterator<Element> configIt = config.getDescendants(new ElementNameFilter("children")); configIt.hasNext();) {
			Element children = configIt.next();
			//it is the node we were looking for.
			if(id.equals(children.getAttributeValue("origin"))) {
				String state = children.getAttributeValue("state");
				if(state != null) {
					if (state.endsWith("_SELECTED")) {
						return Constants.SELECTED;
					}
					else if(state.endsWith("_ELIMINATED")) {
						return Constants.ELIMINATED;
					}
				}
				else {
					return Constants.UNDEFINED;
				}					
			}
		}
		return null;
	}

	/**
	 * This method returns the state associated with a a given feature.  
	 * @param config the configuration {@link Element} that is to be parsed.
	 * @param id the ID of the desired feature.
	 * @param min the minimum cardinality of the feature 
	 * @param max the maximum cardinality of the feature
	 * @return A {@link String} with the state. <i>selected</i> if the feature is selected, <i>eliminated</i> if the feature is eliminated or null if no feature was found.
	 */
	private String getFmpFeatureState(Element config, String id, int min, int max) {
		//if feature as min=1 and max=1 then it will be selected automatically.
		if(min == 1 && max == 1) {
			return Constants.SELECTED;
		}
		return getFmpFeatureStateAux(config, id);
	}

	/**
	 * This method searches for the propertiesID of a feature given it's ID.
	 * @param id the ID of the feature to be found.
	 * @return the propertiesID of the feature.
	 */
	private String getPropertiesId(String id) {
		Element model = document.getRootElement().getChild("model");
		for(Iterator<Element> childrenIt = model.getDescendants(new ElementNameFilter("children")); childrenIt.hasNext();) {
			Element children = childrenIt.next();
			if(id.equalsIgnoreCase(children.getAttributeValue("id",getFmpNamespace("xmi")))) {
				return children.getAttributeValue("properties");
			}
		}
		return null;
	}


	/**
	 * This function returns a list lists containing all the configs present in a FMP model.
	 * @return All the configs, or an empty list if there are none.
	 */
	public SortedMap<String, List<ConfigFmpElement>> getFmpConfigurations() {
		SortedMap<String, List<ConfigFmpElement>> map = new TreeMap<String, List<ConfigFmpElement>>();
		Element children = document.getRootElement().getChild("model").getChild("children");
		List<Element> configs = children.getChildren("configurations");
		for(int i=0; i<configs.size(); i++) {
			Element config = configs.get(i);
			List<ConfigFmpElement> elems = new ArrayList<ConfigFmpElement>(); 
			for(Iterator<Element> childrenIt = config.getDescendants(new ElementNameFilter("children")); childrenIt.hasNext();) {
				Element current = childrenIt.next();
				String id = current.getAttributeValue("origin");
				String propertiesId = getPropertiesId(id);
				ConfigFmpElement feature = new ConfigFmpElement(id,getFmpFeatureProperties(propertiesId));
				String state = getFmpFeatureState(config,id,feature.getProperties().getMin(),feature.getProperties().getMax());
				if (state != null) {
					feature.setState(state);
				}
				elems.add(feature);
			}
			String key = "Configuration " + (i+1);
			map.put(key, elems);
		}
		return map;
	}
}