package org.ample.vml.vml4req.editor.actions;

import java.io.IOException;
import java.io.StringBufferInputStream;
import java.lang.reflect.InvocationTargetException;
import java.net.URL;
import java.util.Iterator;

import org.ample.vmlstar.util.PluginHelper;
import org.ample.vmlstar.util.TraceHelper;
import org.ample.vmlstar.util.WorkflowUtil;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IFolder;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.FileLocator;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.SubMonitor;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.resource.impl.ResourceSetImpl;
import org.eclipse.jdt.core.IClasspathEntry;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.JavaCore;
import org.eclipse.jdt.internal.core.ClasspathEntry;
import org.eclipse.jface.action.IAction;
import org.eclipse.jface.operation.IRunnableWithProgress;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.ui.IActionDelegate;
import org.eclipse.ui.PlatformUI;

import org.osgi.framework.BundleException;

/**
 * Action to trigger compilation of a vml4req script. Compilation of a vml4req 
 * script produces openarchitectureWare model transformations that can 
 * manipulate target models as indicated by a specific product configuration.
 *  
 * @see IActionDelegate
 */
public class CompileVml4reqAction implements IActionDelegate {

	/**
	 * The current selection in the workbench.
	 */
	private ISelection selection;

	/**
	 * Selection in the workbench has changed.
	 * 
	 * @see IActionDelegate#selectionChanged
	 */
	public void selectionChanged(IAction action, ISelection selection) {
		this.selection = selection;
	}

	/**
	 * Called when the action was selected. For every file in the selection, we
	 * invoke process().
	 */
	public void run(IAction action) {
		if (selection instanceof IStructuredSelection) {
			for (Iterator<?> i = ((IStructuredSelection) selection).iterator(); i
					.hasNext();) {
				Object o = i.next();
				if (o instanceof IFile) {
					IFile file = (IFile) o;
					// TODO Possibly check for file extension etc.
					process(file);
				}
			}
		}
	}

	/**
	 * Invoked for every file for which the action has been invoked. Triggers a
	 * separate job to get us out of the UI thread as quickly as possible. Also,
	 * this gives us a progress monitor to use for reporting any progress we
	 * might be making.
	 * 
	 * @param file
	 */
	private void process(final IFile file) {
		try {
			IRunnableWithProgress runnable = new IRunnableWithProgress() {
				public void run(IProgressMonitor monitor)
						throws InvocationTargetException, InterruptedException {
					try {
						doProcess(file, monitor);
					} catch (IOException ioe) {
						ioe.printStackTrace();
					} catch (CoreException e) {
						e.printStackTrace();
					}
				}
			};
			PlatformUI.getWorkbench().getProgressService().busyCursorWhile(
					runnable);
		} catch (InvocationTargetException e) {
			e.getTargetException().printStackTrace();
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
	}

	/**
	 * The actual action implementation. Here we invoke the compile.oaw workflow
	 * that does the actual compilation.
	 * @throws IOException 
	 * @throws CoreException 
	 */
	private void doProcess(IFile file, IProgressMonitor monitor) throws IOException, CoreException {
		try {
			// First ensure the target project is indeed an oAW project
			if (!WorkflowUtil.ensureIsOawProject(file.getProject(), null, monitor)) {
				return;
			}
			// Also need to ensure that stuff required by tracing is available
			TraceHelper.updateProjectClasspath(file.getProject(), monitor);
			// Also need to ensure that we can access oAW's UML2 stuff
			//ensureUML2IsAccessible(file.getProject(), monitor);
		
			// Also need to ensure that the generated code will be able to find the
			// extensions it needs...
			SubMonitor progress = SubMonitor.convert(monitor, 6);
			PluginHelper.addPluginDependencies(file.getProject(), true,
					progress, 
					new PluginHelper.PluginDescriptor("unl.vml4req"), 
					new PluginHelper.PluginDescriptor("ca.uwaterloo.gp.fmp"), 
					new PluginHelper.PluginDescriptor("org.openarchitectureware.uml2.adapter"), 
					new PluginHelper.PluginDescriptor("org.openarchitectureware.util.uml2ecore"), 
					//new PluginHelper.PluginDescriptor("net.ample.tracing.core"), No longer including this, as we are using ATF off-line
                    new PluginHelper.PluginDescriptor("org.ample.vmlstar.util"),
                    new PluginHelper.PluginDescriptor("org.ample.vml.vml4req"),
					new PluginHelper.PluginDescriptor("org.ample.vml.vml4req.generator")/*,
				Removed the following, as they are UML-specific. They should really be 
				explicitly mentioned in the language descriptor.
					new PluginHelper.PluginDescriptor("org.openarchitectureware.uml2.adapter"),
					new PluginHelper.PluginDescriptor("org.openarchitectureware.util.uml2ecore")*/);
		} catch (BundleException e) {
			e.printStackTrace();
		}
		
		// Copy libraries contained in any of the bundles we depend on and make
		// them direct entries on our classpath.
		// This is stupid, but needed so that oAW will find them and be able to
		// use them...
		WorkflowUtil.addRequiredLibraries(file.getProject(), monitor, new String[] {"unl.vml4req", "ca.uwaterloo.gp.fmp", "org.openarchitectureware.uml2.adapter", "org.openarchitectureware.util.uml2ecore"});		
		
		// Obtain the URL of the workflow we are going to use. This is in the
		// generator project.
		// Note that using FileLocator for this doesn't work as it will produce different 
		// results depending on whether we are running from source or from a deployed JAR file.
		URL urlToWf = getClass().getResource("/org/ample/vml/vml4req/compile.oaw");
		/*
		FileLocator
				.toFileURL(new URL(
						"platform:/plugin/org.ample.vml.vml4req.generator/org/ample/vml/vml4req/compile.oaw"));*/
		if (urlToWf == null) {
			System.out
					.println("The Vml4req editor is not able to find the compilation code.");
			System.out
					.println("Contact pablo@lcc.uma.es and report about the bug 'compile.oaw' could not be found'.");
			return;
		}
		// Copy workflow
		IFolder srcGenFolder = file.getProject().getFolder("/src-gen");
		if (!srcGenFolder.exists()) {
			// This is a problem.
			throw new IOException(
					"Folder /src-gen/ does not exist. This is unexpected!");
		}
		IFile fTarget = srcGenFolder.getFile("compile.oaw");
		if (fTarget.exists()) {
			// Clean it
			fTarget.delete(true, monitor);
		}
		// Copy it
		fTarget.create(urlToWf.openStream(), true, monitor);

		// Generate properties file
		StringBuffer sbProperties = new StringBuffer();
		sbProperties.append("vmlFile=").append(file.getLocation()).append('\n');
		sbProperties.append("targetDir=").append(file.getProject().getFolder("src-gen").getLocation()).append('\n');
		// Determine target model URI. This is stored in the vml4req file, so we only need to look there...
		// When opening the file as a resource, it will be parsed automatically, as xText has created a special resource adapter for it.
		ResourceSet rs = new ResourceSetImpl();		
		EObject eoVml4reqModel = rs.getResource(URI.createURI(file.getLocationURI().toString()), true).getContents().get(0);
		EPackage epVml4req = EPackage.Registry.INSTANCE.getEPackage("http://www.ample.org/vml/vml4req");
		if (epVml4req == null) {
			// This shouldn't happen, as the metamodel should have been registered when parsing the vml4req file
			System.err.println("Couldn't access Vml4req metamodel from registry!");
			return;
		}
		EClass ecModel = (EClass) epVml4req.getEClassifier ("Model");
		EStructuralFeature esfCoreModel = ecModel.getEStructuralFeature("coreModel");
		EObject eoCoreModel = (EObject) eoVml4reqModel.eGet(esfCoreModel);
		EClass ecCoreModelImport = (EClass) epVml4req.getEClassifier ("CoreModelImport");
		EStructuralFeature esfCoreURI = ecCoreModelImport.getEStructuralFeature("coreURI");
		sbProperties.append("targetModel=").append(eoCoreModel.eGet(esfCoreURI)).append('\n');
		
		IFile propFile = srcGenFolder.getFile("compileConfig.properties");
		if (propFile.exists()) {
			propFile.delete(true, monitor);
		}
		propFile.create(new StringBufferInputStream(sbProperties.toString()), true, monitor);
		
		// Invoke workflow 
		IFile fWorkflow = srcGenFolder.getFile("compile.oaw");
		WorkflowUtil.executeWorkflow("Compiling " + file.getName(), fWorkflow);
	}

	private void ensureUML2IsAccessible(IProject project,
			IProgressMonitor monitor) throws CoreException, IOException {
		IFolder fResources = project.getFolder("libs");
		IFile fTgtJar = fResources.getFile("uml2helper_1.0.0.jar");
		if (fTgtJar.exists()) {
			return;
		}

		if (!fResources.exists()) {
			fResources.create(true, true, monitor);
		}

		URL urlSrcJar = FileLocator
				.toFileURL(new URL(
						"platform:/plugin/org.ample.vml.vml4req.editor/resources/uml2helper_1.0.0.jar"));
		fTgtJar.create(urlSrcJar.openStream(), true, monitor);

		IJavaProject jp = JavaCore.create(project);
		IClasspathEntry[] origClasspath = jp.getRawClasspath();
		IClasspathEntry[] newClasspath = new ClasspathEntry[origClasspath.length + 1];
		System.arraycopy(origClasspath, 0, newClasspath, 1,
				origClasspath.length);
		newClasspath[0] = JavaCore.newLibraryEntry(fTgtJar.getLocation(), null,
				null);
		jp.setRawClasspath(newClasspath, monitor);
	}
}