package org.ample.vml.vml4req.editor.actions;

import java.io.StringBufferInputStream;
import java.io.IOException;
import java.net.URL;
import java.util.Iterator;

import org.ample.vmlstar.util.WorkflowUtil;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IFolder;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.FileLocator;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.jobs.Job;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.resource.impl.ResourceSetImpl;
import org.eclipse.jface.action.IAction;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.swt.SWT;
import org.eclipse.swt.widgets.FileDialog;
import org.eclipse.ui.IActionDelegate;
import org.eclipse.ui.PlatformUI;

/**
 * Action to trigger product configuration based on a vml4req script.
 * Configuration based on a vml4req script takes a product configuration file
 * and triggers the workflow previously generated by compiling the vml4req
 * script. This produces the configuration-specific models.
 * 
 * @see IActionDelegate
 */
public class ConfigureVml4reqAction implements IActionDelegate {

	/**
	 * The current selection in the workbench.
	 */
	private ISelection selection;

	/**
	 * Selection in the workbench has changed.
	 * 
	 * @see IActionDelegate#selectionChanged
	 */
	public void selectionChanged(IAction action, ISelection selection) {
		this.selection = selection;
	}

	/**
	 * Called when the action was selected. For every file in the selection, we
	 * invoke process().
	 */
	public void run(IAction action) {
		if (selection instanceof IStructuredSelection) {
			for (Iterator<?> i = ((IStructuredSelection) selection).iterator(); i
					.hasNext();) {
				Object o = i.next();
				if (o instanceof IFile) {
					IFile file = (IFile) o;
					// TODO Possibly check for file extension etc.
					process(file);
				}
			}
		}
	}

	/**
	 * Invoked for every file for which the action has been invoked. Triggers a
	 * separate job to get us out of the UI thread as quickly as possible. Also,
	 * this gives us a progress monitor to use for reporting any progress we
	 * might be making.
	 * 
	 * @param file
	 */
	private void process(final IFile file) {
		// Ask the user for the configuration file
		FileDialog openDialog = new FileDialog(PlatformUI.getWorkbench()
				.getActiveWorkbenchWindow().getShell(), SWT.OPEN
				| SWT.PRIMARY_MODAL);
		// openDialog.setFilterExtensions(filters);
		openDialog.setFilterPath(file.getProject().getLocation().toOSString());
		openDialog.setText("Specify a configuration file");
		final String sConfigurationFile = openDialog.open();
		if (sConfigurationFile == null) {
			// User cancelled or an error occurred.
			return;
		}

		// Do not use busyCursorWhile here, as this blocks the UI thread, 
		// which in turn seems to block our ability to trigger any launch 
		// configurations further down.
		new Job("Producing target models for configuration "
				+ sConfigurationFile + " based on " + file.getName()) {

			@Override
			protected IStatus run(IProgressMonitor monitor) {
				try {
					doProcess(file, sConfigurationFile, monitor);
					return Status.OK_STATUS;
				} catch (IOException ioe) {
					ioe.printStackTrace();
					return new Status(IStatus.ERROR,
							"org.ample.vml.vml4req.editor",
							"An error occurred: " + ioe.getLocalizedMessage(),
							ioe);
				} catch (CoreException ce) {
					ce.printStackTrace();
					return ce.getStatus();
				}
			}

		}.schedule();
	}

	/**
	 * The actual action implementation. Here we invoke the workflow previously
	 * generated that does the actual configuration.
	 * 
	 * @throws IOException
	 * @throws CoreException 
	 */
	private void doProcess(IFile file, String sConfigurationFile,
			IProgressMonitor monitor) throws IOException, CoreException {
		// First ensure the target project is indeed an oAW project
		if (!WorkflowUtil.ensureIsOawProject(file.getProject(), null, monitor)) {
			return;
		}
		/*
		 * Steps to success:
		 * 
		 * 1. Ask the user for the configuration file. Already done in process, as it still needs to be in the UI thread.
		 * 
		 * 2. Copy the standard workflows from the plugin into the workspace.
		 * 
		 * 3. Create a property file sitting right next to the workflow file and
		 * fill in the path to the configuration file
		 * 
		 * TODO: 3a. Enter other interesting stuff into the property file, such
		 * as target dir and vml4req URL.
		 * 
		 * 4. Use the WorkflowRunnerHelper to actually run the workflow.
		 */
		// Copy standard workflows
		String[] saFilesToCopy = { "createExecutables.oaw" };
		IFolder srcGenFolder = file.getProject().getFolder("/src-gen");
		if (!srcGenFolder.exists()) {
			// This is a problem. It probably means they haven't compiled the
			// vml4req script first
			throw new IOException(
					"Folder /src-gen/ does not exist. Please first compile "
							+ file.getName());
		}
		for (String sCurrentFile : saFilesToCopy) {
		    /*
		     * Get the source URL. Note we cannot use FileLocator and platform protocol URLs here, 
		     * as they will behave differently for deployed plugins and when running from source. 
		     */
			URL srcURL = getClass().getResource("/org/ample/vml/vml4req/" + sCurrentFile); 
			/*
			FileLocator.toFileURL(new URL(
					"platform:/plugin/org.ample.vml.vml4req.generator/org/ample/vml/vml4req/"
							+ sCurrentFile));*/
			IFile fTarget = srcGenFolder.getFile(sCurrentFile);
			if (fTarget.exists()) {
				// Clean it
				fTarget.delete(true, monitor);
			}
			// Copy it
			fTarget.create(srcURL.openStream(), true, monitor);
		}

		// Generate properties file
		StringBuffer sbProperties = new StringBuffer();
		sbProperties.append("configFile=").append(URI.createFileURI(sConfigurationFile).toString()).append('\n');
		sbProperties.append("vmlFile=").append(file.getLocationURI().toString()).append('\n');
		sbProperties.append("targetDir=/").append(file.getProject().getLocation().toString()).append("/src-gen/").append('\n');
		// Determine target model URI. This is stored in the vml4req file, so we only need to look there...
		// When opening the file as a resource, it will be parsed automatically, as xText has created a special resource adapter for it.
		ResourceSet rs = new ResourceSetImpl();		
		EObject eoVml4reqModel = rs.getResource(URI.createURI(file.getLocationURI().toString()), true).getContents().get(0);
		EPackage epVml4req = EPackage.Registry.INSTANCE.getEPackage("http://www.ample.org/vml/vml4req");
		if (epVml4req == null) {
			// This shouldn't happen, as the metamodel should have been registered when parsing the vml4req file
			System.err.println("Couldn't access Vml4req metamodel from registry!");
			return;
		}
		EClass ecModel = (EClass) epVml4req.getEClassifier ("Model");
		EStructuralFeature esfCoreModel = ecModel.getEStructuralFeature("coreModel");
		EObject eoCoreModel = (EObject) eoVml4reqModel.eGet(esfCoreModel);
		EClass ecCoreModelImport = (EClass) epVml4req.getEClassifier ("CoreModelImport");
		EStructuralFeature esfCoreURI = ecCoreModelImport.getEStructuralFeature("coreURI");
		sbProperties.append("targetModel=").append(eoCoreModel.eGet(esfCoreURI)).append('\n');
		IFile propFile = srcGenFolder.getFile("configuration.properties");
		if (propFile.exists()) {
			propFile.delete(true, monitor);
		}
		propFile.create(new StringBufferInputStream(sbProperties.toString()), true, monitor);

		// Invoke 1st workflow generating negative and positive workflows 
		IFile fWorkflow = srcGenFolder.getFile("createExecutables.oaw");
		WorkflowUtil.executeWorkflow("Configuring based on " + file.getName(),
				fWorkflow);
		
		// Invoke the generated workflow
		fWorkflow = srcGenFolder.getFile("configuration.oaw");
		WorkflowUtil.executeWorkflow("Configuring based on " + file.getName(),
				fWorkflow);
	}
}