package org.ample.vmlstar.langinst.popup.actions;

import hub.sam.tef.Utilities;
import hub.sam.tef.modelcreating.IModelCreatingContext;
import hub.sam.tef.modelcreating.ModelCreating;
import hub.sam.tef.modelcreating.ModelCreatingContext;
import hub.sam.tef.modelcreating.ModelCreatingException;
import hub.sam.tef.semantics.AbstractError;
import hub.sam.tef.semantics.DefaultIdentificationScheme;
import hub.sam.tef.semantics.DefaultSemanticsProvider;
import hub.sam.tef.tsl.Syntax;
import hub.sam.tef.tsl.TslException;

import java.io.BufferedReader;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.lang.reflect.InvocationTargetException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.jar.Manifest;

import org.ample.vmlLangInst.EvaluationAspect;
import org.ample.vmlLangInst.LangInstModel;
import org.ample.vmlLangInst.util.EglHelper;
import org.ample.vmlLangInst.util.LIMResourceManager;
import org.ample.vmlstar.langinst.Activator;
import org.ample.vmlstar.langinst.ILangInstConstants;
import org.ample.vmlstar.util.PluginHelper;
import org.ample.vmlstar.util.WorkflowUtil;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IFolder;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.FileLocator;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.Path;
import org.eclipse.core.runtime.Platform;
import org.eclipse.core.runtime.SubMonitor;
import org.eclipse.core.runtime.jobs.Job;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.resource.impl.ResourceImpl;
import org.eclipse.epsilon.egl.exceptions.EglRuntimeException;
import org.eclipse.jface.action.IAction;
import org.eclipse.jface.operation.IRunnableWithProgress;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.IActionDelegate;
import org.eclipse.ui.IObjectActionDelegate;
import org.eclipse.ui.IWorkbenchPart;
import org.eclipse.ui.PlatformUI;
import org.openarchitectureware.wizards.EclipseHelper;
import org.osgi.framework.BundleException;

public class GenerateVMLInstanceAction implements IObjectActionDelegate,
		ILangInstConstants {

	private Shell shell;
	private ISelection selection;
	private LangInstModel lim;
	private IProject vmlProject;
	private IProject vmlEditorProject;
	private IProject vmlGeneratorProject;
	private String srcProjectName;

	/**
	 * We just remember a shell, which is useful in showing message boxes etc.
	 * 
	 * @see IObjectActionDelegate#setActivePart(IAction, IWorkbenchPart)
	 */
	public void setActivePart(IAction action, IWorkbenchPart targetPart) {
		shell = targetPart.getSite().getShell();
	}

	/**
	 * We just remember the current selection.
	 * 
	 * @see IActionDelegate#selectionChanged(IAction, ISelection)
	 */
	public void selectionChanged(IAction action, ISelection selection) {
		this.selection = selection;
	}

	/**
	 * @see IActionDelegate#run(IAction)
	 */
	public void run(IAction action) {
		if (selection instanceof IStructuredSelection) {
			for (Iterator<?> i = ((IStructuredSelection) selection).iterator(); i
					.hasNext();) {
				Object o = i.next();
				if (o instanceof IFile) {
					IFile file = (IFile) o;
					if (file.getFileExtension().startsWith("vmlinst"))
						process(file);
				}
			}
		}
	}

	/**
	 * Start the processing of this request as a scheduled job in the workbench.
	 * 
	 * @param file
	 */
	private void process(final IFile file) {
		try {
			IRunnableWithProgress runnable = new IRunnableWithProgress() {
				public void run(IProgressMonitor monitor)
						throws InvocationTargetException, InterruptedException {
					doProcess(file, monitor);
				}
			};
			PlatformUI.getWorkbench().getProgressService().busyCursorWhile(
					runnable);
		} catch (InvocationTargetException e) {
			e.getTargetException().printStackTrace();
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
	}

	/**
	 * This is the actual handler for one file. As this uses some member
	 * variables for storing intermediary references, it needs to be thread
	 * safe.
	 * 
	 * @param file
	 * @param monitor
	 */
	private synchronized void doProcess(IFile file, IProgressMonitor monitor) {
		try {
			SubMonitor progress = SubMonitor.convert(monitor, 100);

			srcProjectName = file.getFullPath().segment(0);

			lim = parse(file, progress.newChild(20));

			createProject(lim, progress.newChild(5));

			generateExtensions(lim, progress.newChild(2));

			generateChecks(lim, progress.newChild(2));

			generateLinking(lim, progress.newChild(2));

			generatePropertySettings(lim, progress.newChild(3));

			generateEvaluationAspects(lim, srcProjectName, progress);

			generatePackageDependencies(lim, progress);

			setProjectVersion(vmlProject, lim.getVersion(), monitor);
			setProjectVersion(vmlEditorProject, lim.getVersion(), monitor);
			setProjectVersion(vmlGeneratorProject, lim.getVersion(), monitor);
			
			LIMResourceManager.removeResourceFor(lim);
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	/**
	 * Parse the given file into a {@see LangInstModel}.
	 * 
	 * @param file
	 * @return
	 * @throws VMLInstParsingException
	 */
	private LangInstModel parse(IFile file, SubMonitor progress)
			throws VMLInstParsingException {
		try {
			String text;
			try {
				BufferedReader r = new BufferedReader(new InputStreamReader(
						file.getContents()));
				StringBuffer sbText = new StringBuffer();
				String sLine;
				do {
					sLine = r.readLine();
					if (sLine != null) {
						sbText.append(sLine);
						sbText.append("\n");
					}
				} while (sLine != null);
				text = sbText.toString();
			} catch (Exception e) {
				throw new VMLInstParsingException(e.getMessage(), e);
			}

			IModelCreatingContext context = new ModelCreatingContext(
					epaMetamodels, new DefaultSemanticsProvider(
							DefaultIdentificationScheme.INSTANCE),
					new ResourceImpl(), text);

			EObject eoLangInstModel;

			try {
				Syntax sSyntax = Utilities.loadSyntaxDescription(bSyntaxBundle,
						sSyntaxPath, epaMetamodels);

				eoLangInstModel = ModelCreating.createModelFromText(context,
						sSyntax);
			} catch (TslException te) {
				throw new VMLInstParsingException(te.getMessage(), te);
			} catch (ModelCreatingException e) {
				throw new VMLInstParsingException(e.getMessage(), e);
			}

			// Check for any errors that might have occurred...
			VMLInstParsingException except = null;
			for (AbstractError error : context.getErrors()) {
				if (except == null) {
					except = new VMLInstParsingException();
				}

				except.add(error, context);
			}
			if (except != null) {
				throw except;
			}

			if ((eoLangInstModel != null)
					&& (eoLangInstModel instanceof LangInstModel)) {
				return (LangInstModel) eoLangInstModel;
			} else {
				throw new VMLInstParsingException(
						"Parsing did not result in a language instance model.");
			}
		} finally {
			progress.done();
		}
	}

	/**
	 * This code copied from oAW code, as it was not exported from there...
	 * 
	 * @author zschaler
	 */
	private class ProjectInfo {
		private String projectName;
		private String languageName;
		private String fileExtension;
		private String nsURI;
		private String basePackage;
		private boolean createGeneratorProject = false;

		public boolean isCreateGeneratorProject() {
			return createGeneratorProject;
		}

		public void setCreateGeneratorProject(boolean createGeneratorProject) {
			this.createGeneratorProject = createGeneratorProject;
		}

		public String getBasePackage() {
			return basePackage;
		}

		public void setBasePackage(String basePackage) {
			this.basePackage = basePackage;
		}

		public String getFileExtension() {
			return fileExtension;
		}

		public void setFileExtension(String fileExtension) {
			this.fileExtension = fileExtension;
		}

		public String getLanguageName() {
			return languageName;
		}

		public void setLanguageName(String languageName) {
			this.languageName = languageName;
		}

		public String getNsURI() {
			return nsURI;
		}

		public void setNsURI(String nsURI) {
			this.nsURI = nsURI;
		}

		public String getProjectName() {
			return projectName;
		}

		public void setProjectName(String projectName) {
			this.projectName = projectName;
		}

		public String getGrammarName() {
			return getLanguageName().replaceAll("\\s", "_") + ".xtxt";
		}

		public String getBasePath() {
			return getBasePackage().replaceAll("\\.", "/");
		}		
	}

	/**
	 * Creates the initial xText projects.
	 * 
	 * TODO We should get rid of this and use doCreateProject directly. This
	 * would allow us to also throw out ProjectInfo etc.
	 * 
	 * @param lim
	 * @param progress
	 * @throws CoreException
	 * @throws EglRuntimeException
	 * @throws IOException
	 */
	private void createProject(LangInstModel lim, SubMonitor progress)
			throws CoreException, EglRuntimeException, IOException {
		/*
		 * I found out that oAW comes with its own class for creating such
		 * projects. So, lets just use that code
		 */
		ProjectInfo pi = new ProjectInfo();
		pi.setBasePackage("org.ample.vml." + lim.getName());
		pi.setCreateGeneratorProject(true);
		pi.setFileExtension(lim.getName());
		pi.setLanguageName(lim.getName());
		pi.setNsURI("http://www.ample.org/vml/" + lim.getName());
		pi.setProjectName("org.ample.vml." + lim.getName());

		doCreateProject(pi, progress);
	}

	private static final String PACKAGE_XTEXT_EDITOR_BASE = "org.openarchitectureware.xtext.editor.base";
	private static final String PACKAGE_XTEXT_CORE = "org.openarchitectureware.xtext.core";
	private static final String PACKAGE_XTEXT_CORE_BASE = "org.openarchitectureware.xtext.core.base";
	private static final String PACKAGE_XTEXT_GENERATOR = "org.openarchitectureware.xtext.generator";
	private static final String ORG_ANTLR = "org.antlr";

	/**
	 * This code copied and adjusted from
	 * org.openarchitectureware.xtext.ui.wizards.XtextProjectsWizard
	 * .ProjectCreator#execute. Stupidly, the creators of oAW forgot to export
	 * these packages when creating that plugin, so I cannot simply reuse it.
	 * However, as I've had to make a few changes anyway, maybe it is all for
	 * the best this way :-)
	 * 
	 * @param monitor
	 * @throws CoreException
	 * @throws IOException
	 * @throws EglRuntimeException
	 * @throws InvocationTargetException
	 * @throws InterruptedException
	 */
	private void doCreateProject(ProjectInfo pi, final IProgressMonitor monitor)
			throws CoreException, EglRuntimeException, IOException {
		monitor.beginTask("Creating project " + pi.getProjectName(), 2);

		final Set<String> refs = new HashSet<String>();
		refs.add(PACKAGE_XTEXT_CORE_BASE);
		refs.add(PACKAGE_XTEXT_GENERATOR);
		refs.add(PACKAGE_XTEXT_CORE);
		refs.add(ORG_ANTLR);
		List<String> srcfolders = new ArrayList<String>();
		srcfolders.add("src");
		srcfolders.add("src-gen");

		final List<String> exportedPackages = new ArrayList<String>();
		// exportedPackages.add(pi.getBasePackage());
		// exportedPackages.add(pi.getProjectName() + ".parser");

		vmlProject = EclipseHelper.createOAWProject(pi.getProjectName(),
				srcfolders, Collections.<IProject> emptyList(), refs,
				exportedPackages, monitor, shell);
		if (vmlProject == null)
			return;
		monitor.worked(1);
		final IFolder f = (IFolder) vmlProject.findMember("src");
		final String wf = "<workflow>\n   <property file='generate.properties'/>\n"
				+ "   <component file='org/openarchitectureware/xtext/Generator.oaw' inheritAll='true'/>\n"
				+ "</workflow>";
		IFile fWorkflow = EclipseHelper.createFile("generate.oaw", f, wf,
				monitor);

		String props = "workspace.dir=../\n\n";
		props += "core.project.name=" + pi.getProjectName() + "\n";
		props += "grammar=./src/" + pi.getGrammarName() + "\n";
		props += "debug.grammar=true\n";
		props += "language.name=" + pi.getLanguageName() + "\n";
		props += "language.nsURI=" + pi.getNsURI() + "\n";
		props += "language.fileextension=" + pi.getFileExtension() + "\n";
		props += "basepath=" + pi.getBasePath() + "\n";
		props += "generator.project.name=" + pi.getProjectName()
				+ ".generator\n\n";

		props += "generate.standaloneParser=false\n\n";

		props += "#IMPORTANT! When the following flag is activated (i.e. set to true) the generator will overwrite the manifest and plugin resources!\n";
		props += "overwrite.pluginresources=true\n";

		EclipseHelper.createFile("generate.properties", f, props, monitor);

		// Changed below to use a call to a function generating the grammar
		// instead of a hard-coded empty grammar.
		EclipseHelper.createFile(pi.getGrammarName(), f, generateGrammar(),
				monitor);

		// creating editor project
		refs.add(PACKAGE_XTEXT_EDITOR_BASE);
		refs.add(pi.getProjectName());
		vmlEditorProject = EclipseHelper.createOAWProject(pi.getProjectName()
				+ ".editor", Arrays.asList(new String[] { "src", "src-gen",
				"icons" }), Collections.<IProject> emptyList(), refs, null,
				monitor, shell);
		// add default outline icon to the editor projects icons
		// folder
		final IFolder iconsFolder = (IFolder) vmlEditorProject
				.findMember("icons");
		EclipseHelper.createFile("default.gif", iconsFolder, Platform
				.getBundle(PACKAGE_XTEXT_EDITOR_BASE).getEntry(
						"icons/ov_obj.gif"), monitor);

		/*
		 * Commented below: A class was missing, and there seems to be no point
		 * in opening the grammar in my case anyway.
		 * 
		 * monitor.setTaskName("Opening file for editing...");
		 * BasicNewResourceWizard.selectAndReveal(grammar, PlatformUI
		 * .getWorkbench().getActiveWorkbenchWindow());
		 * EclipseHelper.openFileToEdit(shell, grammar); monitor.worked(1);
		 */

		if (pi.isCreateGeneratorProject()) {
			srcfolders = new ArrayList<String>();
			srcfolders.add("src");
			// creating generator project
			vmlGeneratorProject = EclipseHelper.createOAWProject(pi
					.getProjectName()
					+ ".generator", srcfolders, Collections
					.<IProject> emptyList(), refs, null, monitor, shell);

			final StringBuilder wf1 = new StringBuilder();
			wf1.append("<workflow>\n");
			wf1.append("  <property name='modelFile' />\n");
			wf1.append("  <property name='targetDir' value='src-gen/'/>\n");
			wf1
					.append("  <bean class=\"org.eclipse.mwe.emf.StandaloneSetup\">\n");
			wf1.append("     <platformUri value=\"..\"/>\n");
			wf1.append("  </bean>\n");
			wf1.append("  <component file='" + pi.getBasePath()
					+ "/parser/Parser.oaw'>\n");
			wf1.append("     <modelFile value='${modelFile}'/>\n");
			wf1.append("     <outputSlot value='theModel'/>\n");
			wf1.append("  </component>\n");
			wf1
					.append("  <component class='oaw.workflow.common.DirectoryCleaner' directory='${targetDir}'/>\n");
			wf1.append("  <component class='oaw.xpand2.Generator'>\n");
			wf1
					.append("     <metaModel id='mm' class='org.eclipse.m2t.type.emf.EmfRegistryMetaModel'/>\n");
			wf1.append("     <expand value='"
					+ pi.getBasePath().replaceAll("/", "::")
					+ "::Main::main FOR theModel'/>\n");
			wf1.append("     <genPath value='${targetDir}'/>\n");
			wf1.append("  </component>\n");
			wf1.append("</workflow>");
			EclipseHelper.createFile("src/" + pi.getBasePath()
					+ "/generator.oaw", vmlGeneratorProject, wf1.toString(),
					monitor);

			final StringBuilder xpt = new StringBuilder();
			xpt.append("\u00ABIMPORT " + pi.getLanguageName().toLowerCase()
					+ "\u00BB\n\n");

			xpt.append("\u00ABREM\u00BB\n");
			xpt
					.append("   This is the 'main' template. \n   You should replace the type 'Object' with your respective meta type.\n");
			xpt.append("\u00ABENDREM\u00BB\n");
			xpt.append("\u00ABDEFINE main FOR Object\u00BB\n   \n");
			xpt.append("\u00ABENDDEFINE\u00BB");

			EclipseHelper.createFile("src/" + pi.getBasePath() + "/Main.xpt",
					vmlGeneratorProject, xpt.toString(),
					EclipseHelper.ISO_8859_1, monitor);
			EclipseHelper
					.createFile(
							"src/HelloPDE.java",
							vmlGeneratorProject,
							"/* Dummy class for PDE (won't export this plugin otherwise:-() */\nclass HelloPDE {}",
							monitor);
		}

		/*
		 * ADDED: Execute the parser-generation workflow right away...
		 */
		String sLaunchName = "generate.oaw - invoked for " + lim.getName();
		WorkflowUtil.executeWorkflow(sLaunchName, fWorkflow);
		
		/*
		 * ADDED: Copy uml2helper_1.0.0.jar to enable UML2 usage in VML projects
		 */
		IFolder fResources = vmlEditorProject.getFolder("resources");
		if (!fResources.exists()) {
			fResources.create(true, true, monitor);
		}
		URL urlSrcJar = FileLocator.toFileURL(new URL("platform:/plugin/org.ample.vmlstar.langinst/resources/uml2helper_1.0.0.jar"));
		IFile fTgtJar = fResources.getFile("uml2helper_1.0.0.jar");
		fTgtJar.create(urlSrcJar.openStream(), true, monitor);
	}

	/**
	 * Set the plugin version of the given project.
	 * 
	 * @param project
	 * @param version
	 * @throws IOException 
	 * @throws CoreException 
	 */
	private void setProjectVersion(IProject project, String version, IProgressMonitor monitor) throws IOException, CoreException {
		IFile manifestFile = project.getFile("META-INF/MANIFEST.MF");
		if (!manifestFile.exists()) {
			throw new IOException("MANIFEST.MF does not yet exist...");
		}
		Manifest mf = new Manifest(manifestFile.getContents());
		
		// Fix version!
		mf.getMainAttributes().putValue("Bundle-Version", version);
		
		Job[] autoBuildJobs;
		Job[] manBuildJobs;
		do {
			autoBuildJobs = Job.getJobManager().find(
					ResourcesPlugin.FAMILY_AUTO_BUILD);
			manBuildJobs = Job.getJobManager().find(
					ResourcesPlugin.FAMILY_MANUAL_BUILD);
		} while (autoBuildJobs.length > 0 || manBuildJobs.length > 0);

		FileOutputStream fos = new FileOutputStream(manifestFile.getLocation()
				.toOSString());
		mf.write(fos);
		fos.close();
		manifestFile.refreshLocal(1, monitor);
	}

	/**
	 * Generate the xText grammar for the VML instance.
	 * 
	 * @param pi
	 * @param monitor
	 * @return
	 * @throws EglRuntimeException
	 * @throws IOException
	 */
	private String generateGrammar() throws EglRuntimeException, IOException {
		URL eglURL = FileLocator.toFileURL(FileLocator.find(Activator
				.getDefault().getBundle(), new Path(
				"transformations/LIM2xText.egl"), null));
		return EglHelper.generateToString(eglURL, lim);
	}

	/**
	 * Regenerate the extensions.ext file in the vml project to add a few useful
	 * standard extensions.
	 * 
	 * TODO: Later we may want to have this enhanced by other extensions
	 * provided by other aspects.
	 * 
	 * @param lim
	 * @param progress
	 */
	private void generateExtensions(LangInstModel lim, SubMonitor progress)
			throws IOException, EglRuntimeException, CoreException {
		URL eglURL = FileLocator.toFileURL(FileLocator.find(Activator
				.getDefault().getBundle(), new Path(
				"transformations/LIM2EXT.egl"), null));
		IFolder extFolder = vmlProject.getFolder(new Path(
				"/src/org/ample/vml/" + lim.getName()));
		IFile extFile = extFolder.getFile("Extensions.ext");
		EglHelper.generateToFile(eglURL, lim, extFile, progress);
		
		eglURL = FileLocator.toFileURL(FileLocator.find(Activator
				.getDefault().getBundle(), new Path(
				"transformations/LIM2ComparatorUtilJAVA.egl"), null));
		IFolder utilFolder = vmlProject.getFolder(new Path(
				"/src/org/ample/vml/" + lim.getName() + "/util"));
		if (!utilFolder.exists()) {
			utilFolder.create(true, true, progress);
		}
		IFile utilFile = utilFolder.getFile("ComparatorUtil.java");
		EglHelper.generateToFile(eglURL, lim, utilFile, progress);
		
		// Make sure the package is exported correctly.
		PluginHelper.addExportedPackage(vmlProject, "org.ample.vml." + lim.getName() + ".util", progress);
	}

	/**
	 * Regenerate the checks.chk file in the vml project to add a few checks
	 * that are common to all VML languages.
	 * 
	 * TODO: Later we would want to place this into an evaluation aspect so that
	 * we could have language instance specific checks, too.
	 * 
	 * @param lim
	 * @param progress
	 * @throws IOException
	 * @throws EglRuntimeException
	 * @throws CoreException
	 */
	private void generateChecks(LangInstModel lim, SubMonitor progress)
			throws IOException, EglRuntimeException, CoreException {
		URL eglURL = FileLocator.toFileURL(FileLocator.find(Activator
				.getDefault().getBundle(), new Path(
				"transformations/LIM2CHK.egl"), null));
		IFolder checksFolder = vmlProject.getFolder(new Path(
				"/src/org/ample/vml/" + lim.getName()));
		IFile checksFile = checksFolder.getFile("Checks.chk");
		EglHelper.generateToFile(eglURL, lim, checksFile, progress);
	}

	/**
	 * Generate Linking.ext.
	 * 
	 * @param lim2
	 * @param progress
	 * @throws IOException
	 * @throws CoreException
	 * @throws EglRuntimeException
	 */
	private void generateLinking(LangInstModel lim2, SubMonitor progress)
			throws IOException, EglRuntimeException, CoreException {
		URL eglURL = FileLocator.toFileURL(FileLocator.find(Activator
				.getDefault().getBundle(), new Path(
				"transformations/LIM2LinkingEXT.egl"), null));
		IFolder linkingFolder = vmlProject.getFolder(new Path(
				"/src/org/ample/vml/" + lim.getName()));
		IFile linkingFile = linkingFolder.getFile("Linking.ext");
		EglHelper.generateToFile(eglURL, lim, linkingFile, progress);
	}

	/**
	 * Generate any project-specific property settings that may be required. In
	 * particular, oAW seems to require the generator project to have a
	 * reference to the UML2 metamodel contributor in its preferences if UML2 is
	 * the target model.
	 * 
	 * @param lim
	 * @param progress
	 * @throws IOException
	 * @throws CoreException
	 */
	private void generatePropertySettings(LangInstModel lim, SubMonitor progress)
			throws CoreException, IOException {
		URL prefSrcURL = FileLocator.toFileURL(FileLocator.find(Activator
				.getDefault().getBundle(), new Path(
				"resources/org.openarchitectureware.base.prefs"), null));
		copySettings(prefSrcURL, vmlProject, progress.newChild(3));
		copySettings(prefSrcURL, vmlEditorProject, progress.newChild(3));
		copySettings(prefSrcURL, vmlGeneratorProject, progress.newChild(3));
	}

	/**
	 * Copy the indicated settings to the given project
	 * 
	 * @param prefSrcURL
	 * @param project
	 * @param progress
	 * @throws CoreException
	 * @throws IOException
	 */
	private void copySettings(URL prefSrcURL, IProject project,
			SubMonitor progress) throws CoreException, IOException {
		IFolder settingsFolder = project.getFolder(new Path("/.settings/"));
		if (!settingsFolder.exists()) {
			settingsFolder.create(true, true, progress.newChild(1));
		} else {
			progress.worked(1);
		}
		IFile prefTgtFile = settingsFolder
				.getFile("org.openarchitectureware.base.prefs");
		if (prefTgtFile.exists()) {
			prefTgtFile.delete(true, progress.newChild(1));
		} else {
			progress.worked(1);
		}
		prefTgtFile.create(prefSrcURL.openStream(), true, progress.newChild(1));
	}

	/**
	 * Sequentially invoke the generators for all EvaluationAspects, if any.
	 * 
	 * @paran lim
	 * @param srcProjectName
	 * @param monitor
	 */
	private void generateEvaluationAspects(LangInstModel lim,
			String srcProjectName, SubMonitor monitor) {
		SubMonitor progress = monitor.newChild(lim.getAspects().size() + 1);
		try {
			for (EvaluationAspect ea : lim.getAspects()) {
				ea.generate(vmlProject.getName(), progress.newChild(1));
			}
		} finally {
			progress.done();
		}
	}

	/**
	 * Generate the declared package dependencies into all projects created.
	 * 
	 * @param lim
	 * @param progress
	 * @throws BundleException
	 * @throws CoreException
	 * @throws IOException
	 */
	private void generatePackageDependencies(LangInstModel lim, SubMonitor sm)
			throws IOException, CoreException, BundleException {
		SubMonitor progress = sm.newChild(6);
		try {
			PluginHelper.PluginDescriptor[] pdDescriptors = new PluginHelper.PluginDescriptor[lim
					.getReqBundles().size()+1];
			int i = 0;
			for (String sReqBundle : lim.getReqBundles()) {
				pdDescriptors[i++] = new PluginHelper.PluginDescriptor(
						sReqBundle);
			}
			// Enable helper classes for all plugin projects
			pdDescriptors[i++] = new PluginHelper.PluginDescriptor(
					"org.ample.vmlstar.util");

			PluginHelper.addPluginDependencies(vmlProject, true, progress,
					pdDescriptors);
			PluginHelper.addPluginDependencies(vmlEditorProject, true,
					progress, pdDescriptors);
			PluginHelper.addPluginDependencies(vmlGeneratorProject, true,
					progress, pdDescriptors);
		} finally {
			progress.done();
		}
	}
}