package org.ample.vmlstar.util;

import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;

import net.ample.tracing.core.ItemManager;
import net.ample.tracing.core.RepositoryManager;
import net.ample.tracing.core.TraceLink;
import net.ample.tracing.core.TraceLinkType;
import net.ample.tracing.core.TraceWorkspace;
import net.ample.tracing.core.TraceableArtefact;
import net.ample.tracing.core.TraceableArtefactType;
import net.ample.tracing.core.TypeManager;
import net.ample.tracing.core.query.Constraints;
import net.ample.tracing.core.query.Query;
import net.ample.tracing.h2.H2PersistenceManager;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IFolder;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.FileLocator;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.SubMonitor;
import org.eclipse.jdt.core.IClasspathEntry;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.JavaCore;
import org.junit.Assert;
import org.osgi.framework.BundleException;

/**
 * Helper class for supporting tracing in product derivation.
 * 
 * @author zschaler
 */

public class TraceHelper {

	private static RepositoryManager repository_manager;

	private static String repoName = null;
	private static File repoWorkspace = new File("c:/tmp");
	private static File profileLocation = new File("c:/tmp/atf-default.xml");

	/**
	 * This keeps the set of features that we need to consider in any trace
	 * links we're currently generating. Notice, this list currently only keeps
	 * the names of these features.
	 */
	private static List<String> lsCurrentFeatures;
	/**
	 * Keeps the name of the ATF repository to be used for storing the trace
	 * links. If null, no tracing should happen. The whole infrastructure setup
	 * for tracing assumes that all variant blocks are executed strictly
	 * sequentially, so such global-variables are indeed admissible.
	 */
	private static String sATFRepository;

	/**
	 * Update the classpath of the given workspace project by adding the jars
	 * required by TraceHelper. This is a workaround needed to ensure oAW finds
	 * all required classes. It is a bit stupid, but I have failed to find a
	 * better alternative.
	 * 
	 * @param project
	 * @param monitor
	 * @throws CoreException
	 * @throws IOException
	 * @throws MalformedURLException
	 * @throws BundleException
	 */
	public static void updateProjectClasspath(IProject project,
			IProgressMonitor monitor) throws CoreException,
			MalformedURLException, IOException, BundleException {
		IFolder fLibs = project.getFolder("libs");
		if (!fLibs.exists()) {
			fLibs.create(true, true, monitor);
		}

		List<IClasspathEntry> lcpeClasspath = new ArrayList<IClasspathEntry>();

		// TODO: This should be generated by traversing the directory
		final String[] saFilesToCopy = { "commons-collections-2.1.1.jar",
				"commons-lang-2.4.jar", "dom4j-1.6.1.jar", "ehcache-1.2.3.jar",
				"h2-1.1.107.jar", "hibernate3.jar", "jdom.jar", "jta.jar",
				"junit.jar", "net.ample.tracing.h2_0.2.1.200906171621.jar",
				"net.ample.tracing.core_0.2.1.200906171621.jar"
		/*
		 * "antlr-2.7.6.jar", "asm-attrs.jar", "asm.jar", "cglib-2.1.3.jar",
		 * "commons-io-1.4.jar", "commons-logging-1.0.4.jar", "xml-apis.jar",
		 */
		};
		for (String sFileToCopy : saFilesToCopy) {
			IFile fTgtJar = fLibs.getFile(sFileToCopy);
			if (fTgtJar.exists()) {
				continue;
			}

			URL urlSrcJar = FileLocator.toFileURL(new URL(
					"platform:/plugin/org.ample.vmlstar.util/lib/"
							+ sFileToCopy));

			fTgtJar.create(urlSrcJar.openStream(), true, monitor);

			lcpeClasspath.add(JavaCore.newLibraryEntry(fTgtJar.getFullPath(),
					null, null));
		}

		if (!lcpeClasspath.isEmpty()) {
			IJavaProject jp = JavaCore.create(project);
			IClasspathEntry[] origClasspath = jp.getRawClasspath();
			IClasspathEntry[] newClasspath = new IClasspathEntry[origClasspath.length
					+ lcpeClasspath.size()];
			System.arraycopy(origClasspath, 0, newClasspath, lcpeClasspath
					.size(), origClasspath.length);
			for (int i = 0; i < lcpeClasspath.size(); i++) {
				newClasspath[i] = lcpeClasspath.get(i);
			}
			jp.setRawClasspath(newClasspath, monitor);
		}

		// Furthermore, we need to add some other plugins into the dependencies
		PluginHelper.addPluginDependencies(project, true, SubMonitor
				.convert(monitor), new PluginHelper.PluginDescriptor(
				"org.eclipse.emf.query"), new PluginHelper.PluginDescriptor(
				"org.eclipse.emf.teneo.annotations"),
				new PluginHelper.PluginDescriptor(
						"org.eclipse.emf.teneo.hibernate"),
				new PluginHelper.PluginDescriptor(
						"org.eclipse.emf.teneo.hibernate.mapper"),
				new PluginHelper.PluginDescriptor("org.eclipse.emf.teneo"));
	}

	public static void setATFRepository(String sRepository)
			throws CoreException {
		try {
			System.err.println("Asked to trace to <" + sRepository + ">");
			if (sRepository.equals("")) {
				sATFRepository = null;
			} else {
				sATFRepository = sRepository;
				// Connect to the repository
				System.out.println("connecting to repository");
				// TraceHelper.repository_manager =
				// TraceWorkspace.getInstance().getRepositoryManager(
				// sATFRepository );
				// code copied from the example
				repoName = "repo-" + System.currentTimeMillis(); // TODO
				// Steffen:
				// Why is
				// this not
				// simply
				// sATFRepository??
				System.out.println("Creating Repository...");
				repository_manager = TraceWorkspace.getInstance()
						.newRepositoryManager(repoWorkspace, repoName);
				System.out.println("Repository Created");
				repository_manager.createRepository(new H2PersistenceManager());
				repository_manager.initializeRepository(profileLocation);
				// do the test
				Assert.assertTrue(new File(repoWorkspace, repoName).exists());
				// check for initialization
				Assert.assertTrue(repository_manager.getRepository()
						.getArtefactTypes().size() > 0);
				Assert.assertTrue(repository_manager.getRepository()
						.getLinkTypes().size() > 0);
			}
			System.err
					.println("ATF Repository set to <" + sATFRepository + ">");
		} catch (CoreException e) {
			System.out.println("ERROR OCURRED:");
			e.printStackTrace();
			throw e;
		}
	}

	public static void setLinkedFeatures(List<String> lsFeatures) {
		lsCurrentFeatures = new ArrayList<String>(lsFeatures);
	}

	/**
	 * creates an element with the given type in ATF (if not already exists) and
	 * link (with the correct type) the new element with this feature.
	 * 
	 * @param sTarget
	 *            Name of the target element
	 * @param sTargetType
	 *            Type of the target element
	 */
	public static void traceCreated(String sTarget, String sTargetType) {
		System.out.println("The target information is: " + sTarget + " + "
				+ sTargetType);
		if (sATFRepository != null) {
			if (sTarget == null) {
				// Don't trace null objects
				System.err
						.println("Asked to trace creation of null: Not doing anything! "
								+ "Please reconsider the implementation of your creation actions to ensure that they return some element in every case!");
				return;
			}
			System.err.println("Created " + sTarget + " which is a <"
					+ sTargetType + ">.\n"
					+ "\tThe following features were involved: "
					+ lsCurrentFeatures + "\n" + "\tTracing to <"
					+ sATFRepository + ">");

			System.out.println("creating element type in ATF: " + sTargetType);
			TraceableArtefactType elem_type = TraceHelper
					.createElementTypeOnATF(sTargetType.replaceAll(
							"[^a-zA-Z0-9]", ""));
			System.out.println(" element type in ATF: " + sTargetType
					+ "created successfully");

			TraceableArtefact atf_elem = TraceHelper.createElementOnATF(
					sTarget, elem_type);
			TraceableArtefactType feature_type = TraceHelper
					.createElementTypeOnATF("Feature");

			for (int i = 0; i < lsCurrentFeatures.size(); i++) {
				TraceableArtefact atf_feature = TraceHelper.createElementOnATF(
						lsCurrentFeatures.get(i), feature_type);
				// here we need to know if the feature is selected or not...to
				// create the appropriate link type
				TraceHelper.createLinkTypeOnATF("VMLADDLink");
				TraceHelper
						.createLinkOnATF(atf_feature, atf_elem, "VMLADDLink");
				System.out.println("Traced feature " + atf_feature + "to type "
						+ elem_type + " with name" + sTarget);

			}
		} else {
			System.err.println("Not tracing because no repository given...");
		}
	}

	public static void traceRemoved(String sTarget, String sTargetType) {
		if (sATFRepository != null) {
			if (sTarget == null) {
				// Don't trace null objects
				System.err
						.println("Asked to trace removal of null: Not doing anything!");
				return;
			}
			System.err.println("Removed " + sTarget + " which is a <"
					+ sTargetType + ">.\n"
					+ "\tThe following features were involved: "
					+ lsCurrentFeatures + "\n" + "\tTracing to <"
					+ sATFRepository + ">");

			TraceableArtefactType elem_type = TraceHelper
					.createElementTypeOnATF(sTargetType.replaceAll(
							"[^a-zA-Z0-9]", ""));
			TraceableArtefact atf_elem = TraceHelper.createElementOnATF(
					sTarget, elem_type);
			TraceableArtefactType feature_type = TraceHelper
					.createElementTypeOnATF("Feature");

			for (int i = 0; i < lsCurrentFeatures.size(); i++) {
				TraceableArtefact atf_feature = TraceHelper.createElementOnATF(
						lsCurrentFeatures.get(i), feature_type);
				// here we need to know if the feature is selected or not...to
				// create the appropriate link type
				TraceHelper.createLinkTypeOnATF("VMLRemoveLink");
				TraceHelper.createLinkOnATF(atf_feature, atf_elem,
						"VMLRemoveLink");
			}
		} else {
			System.err.println("Not tracing because no repository given...");
		}
	}

	// *** ATF UTILS ***

	// create a new element type if not already exists in the repository

	public static TraceableArtefactType createElementTypeOnATF(String new_type) {
		System.out.println("creating element type " + new_type);
		TypeManager typeMgr = TraceHelper.repository_manager.getTypeManager();
		TraceableArtefactType elemtype = typeMgr
				.findArtefactTypeByName(new_type);
		if (elemtype == null) {
			TraceHelper.repository_manager.getPersistenceManager().begin();
			TraceableArtefactType req_type = typeMgr.createArtefactType(
					new_type, null);
			TraceHelper.repository_manager.getPersistenceManager()
					.add(req_type);
			try {
				TraceHelper.repository_manager.getPersistenceManager().commit();
				System.out.println("element type created" + new_type);
				return req_type;
			} catch (IOException e) {
				e.printStackTrace();
				return null;
			}
		} else {
			System.out.println("element type created" + new_type);
			return elemtype;
		}

	}

	public static void createLinkTypeOnATF(String new_link_type) {
		System.out.println("Creating link type: " + new_link_type);
		TypeManager typeMgr = TraceHelper.repository_manager.getTypeManager();
		if (TraceHelper.repository_manager.getTypeManager().findLinkTypeByName(
				new_link_type) == null) {
			TraceHelper.repository_manager.getPersistenceManager().begin();
			TraceLinkType traceLinkType = typeMgr.createLinkType(new_link_type,
					(TraceLinkType[]) null);
			TraceHelper.repository_manager.getPersistenceManager().add(
					traceLinkType);
			try {
				TraceHelper.repository_manager.getPersistenceManager().commit();
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
		System.out.println("Link Created: " + new_link_type);
	}

	public static TraceableArtefact createElementOnATF(String element_name,
			TraceableArtefactType artifact_type) {
		// adiciona apenas se o elemento no existe..verifica existencia
		Query<TraceableArtefact> q = TraceHelper.repository_manager
				.getQueryManager().queryOnArtefacts();
		q.add(Constraints.and(Constraints.name(element_name), Constraints
				.isOfType(artifact_type)));
		TraceableArtefact element_artefact = q.executeUnique();
		if (element_artefact == null) {
			try {
				if (element_name != null && !(element_name.equals("null"))
						&& !(element_name.equals("")))
					System.out.println("artifacttype: " + artifact_type);
				ItemManager itemManager = TraceHelper.repository_manager
						.getItemManager();
				TraceHelper.repository_manager.getPersistenceManager().begin();
				TraceableArtefact requirement_artefact = itemManager
						.createTraceableArtefact(artifact_type, element_name);
				TraceHelper.repository_manager.getPersistenceManager().add(
						requirement_artefact);
				TraceHelper.repository_manager.getPersistenceManager().commit();
				return requirement_artefact;
			} catch (IOException e) {
				e.printStackTrace();
				return null;
			}
		} else
			return element_artefact;
	}

	public static void createLinkOnATF(TraceableArtefact feature_artefact,
			TraceableArtefact requirement_artefact, String link_type_name) {
		// System.out.println("Creating link between feature " +
		// feature_artefact + "and " + requirement_artefact + "of type" +
		// link_type_name);
		TraceLinkType linkType = TraceHelper.repository_manager
				.getTypeManager().findLinkTypeByName(link_type_name);
		if (!TraceHelper.existsLinkOnATF(feature_artefact,
				requirement_artefact, linkType)) {
			try {
				TraceHelper.repository_manager.getPersistenceManager().begin();
				TraceLink link = TraceHelper.repository_manager
						.getItemManager().createTraceLink(feature_artefact,
								requirement_artefact, linkType);
				TraceHelper.repository_manager.getPersistenceManager()
						.add(link);
				TraceHelper.repository_manager.getPersistenceManager().commit();
				System.out.println("commit sucedido");
			} catch (IOException err) {
				err.printStackTrace();
			}
		}
		System.out.println("Link created sucessfully!");
	}

	public static boolean existsLinkOnATF(TraceableArtefact x,
			TraceableArtefact y, TraceLinkType type) {
		for (TraceLink l : x.getOutgoingLinks()) {
			for (TraceableArtefact a : l.getTargets()) {
				if (a == y && l.getType() == type) {
					return true;
				}
			}
		}
		return false;
	}

	// *** END - ATF UTILS ***

}