/**
 * 
 */
package org.ample.vmlstar.util;

import java.io.IOException;
import java.io.StringBufferInputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;
import java.util.jar.Manifest;

import org.eclipse.core.resources.ICommand;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IFolder;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IProjectDescription;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.FileLocator;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.Path;
import org.eclipse.core.runtime.SubProgressMonitor;
import org.eclipse.core.runtime.jobs.IJobChangeEvent;
import org.eclipse.core.runtime.jobs.IJobChangeListener;
import org.eclipse.core.runtime.jobs.Job;
import org.eclipse.core.runtime.jobs.JobChangeAdapter;
import org.eclipse.debug.core.DebugPlugin;
import org.eclipse.debug.core.ILaunchConfigurationType;
import org.eclipse.debug.core.ILaunchConfigurationWorkingCopy;
import org.eclipse.debug.ui.DebugUITools;
import org.eclipse.jdt.core.IClasspathEntry;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.JavaCore;
import org.eclipse.jdt.launching.IJavaLaunchConfigurationConstants;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.PlatformUI;
import org.openarchitectureware.core.builder.OawBuilder;
import org.openarchitectureware.core.builder.OawNature;
import org.openarchitectureware.eclipse.launch.IWfLaunchConfigurationConstants;

/**
 * A utility class containing methods to easily trigger an oAW workflow from an
 * Eclipse setting.
 * 
 * @author zschaler
 */
public class WorkflowUtil {

	/**
	 * Execute the given workflow. This waits until execution of the workflow
	 * has finished. Unfortunately, I haven't yet figured out how to find out if
	 * the workflow was successful.
	 * 
	 * @param sLaunchName
	 *            a name to be used internally to identify this workflow. Used
	 *            mostly for debugging purposes at the moment.
	 * @param fWorkflow
	 *            an oAW file. Must be in a oAW project in the current
	 *            workspace.
	 * @throws CoreException
	 *             if something goes wrong
	 */
	public static void executeWorkflow(String sLaunchName, IFile fWorkflow)
			throws CoreException {
		/*
		 * Need to do this from a launch configuration rather than invoking
		 * WorkflowRunner directly to make sure that working directories etc.
		 * are set correctly. The attributes set in the launch configuration
		 * below seem to be the minimum required. In particular, we need to set
		 * the java project name to allow the classpath manager to figure out
		 * the correct classpath from the plugin dependencies of the project.
		 * This is necessary so that WorkflowRunner.class will actually be in
		 * that classpath available to be called...
		 */
		ILaunchConfigurationType configType = DebugPlugin
				.getDefault()
				.getLaunchManager()
				.getLaunchConfigurationType(
						IWfLaunchConfigurationConstants.WORKFLOW_LAUNCH_CONFIGURATION_TYPE_ID);
		final ILaunchConfigurationWorkingCopy launchConfig = configType
				.newInstance(null, sLaunchName);
		launchConfig.setAttribute(
				IWfLaunchConfigurationConstants.ATTR_OAW_WORKFLOW_FILE,
				fWorkflow.getFullPath().toString());
		launchConfig.setAttribute(
				IJavaLaunchConfigurationConstants.ATTR_PROJECT_NAME, fWorkflow
						.getProject().getName());

		/*
		 * Wait for build jobs here. For some reason, whenever I am not
		 * debugging this code, DebugUITools run into a NullPointerException
		 * when trying to do the waiting themselves
		 * 
		 * This workaround code was copied from WorkflowLaunchDelegate.launch()
		 * TODO: It still doesn't seem to work always. Maybe this is related to
		 * DebugUITools.launch requesting that it be called from the UI
		 * thread...
		 */
		Job[] autoBuildJobs;
		Job[] manBuildJobs;
		do {
			autoBuildJobs = Job.getJobManager().find(
					ResourcesPlugin.FAMILY_AUTO_BUILD);
			manBuildJobs = Job.getJobManager().find(
					ResourcesPlugin.FAMILY_MANUAL_BUILD);
		} while (autoBuildJobs.length > 0 || manBuildJobs.length > 0);

		// MOVED down to ensure that we will always be able to identify its
		// end...
		// DebugUITools.launch(launchConfig, "run");

		/*
		 * Wait for the triggered jobs to finish. After a number of unsuccessful
		 * tries the hack below seems to work best. Ideally, oAW would place
		 * there refresh job (or all of their jobs for that matter) into a
		 * specific family. Then we could simply join the jobs of that family.
		 * As they don't, we must rely on oAW implementer not changing the name
		 * of the refresh job.
		 */

		// HACK: Wait for the refresh resources job to finish as this seems to
		// be the last thing happening in a workflow run.
		final boolean fLaunchFinished[] = { false };
		IJobChangeListener listener = new JobChangeAdapter() {
			@Override
			public void done(IJobChangeEvent event) {
				if (event.getJob().getName().equals("refresh resources")) {
					synchronized (this) {
						fLaunchFinished[0] = true;
						this.notify();
					}
				}
			}
		};
		Job.getJobManager().addJobChangeListener(listener);

		/*
		 * We only trigger the launch here so that the listener has already been
		 * installed. This way, even for very quick workflows we should be able
		 * to be informed of the end of the refresh job.
		 * 
		 * Also, we defer to the UI thread to do the launch, as DebugUITools
		 * tells us to...
		 */
		// Display.getDefault()
		PlatformUI.getWorkbench().getDisplay().asyncExec(new Runnable() {
			public void run() {
				DebugUITools.launch(launchConfig, "run");
			}
		});

		synchronized (listener) {
			while (!fLaunchFinished[0]) {
				try {
					listener.wait();
				} catch (InterruptedException e) {
				}
			}
		}
		Job.getJobManager().removeJobChangeListener(listener);
	}

	/**
	 * Checks that the project has everything required of an oAW project. If
	 * not, asks the user if it should adjust the project setup. If the user
	 * agrees, the project will be adjusted.
	 * 
	 * @param project
	 * @return false if the project is still not an oAW project
	 * @throws CoreException
	 */
	public static boolean ensureIsOawProject(final IProject project,
			final Shell shell, IProgressMonitor monitor) throws CoreException {
		IProjectDescription projectDescription = project.getDescription();
		if (!projectDescription.hasNature(OawNature.NATURE_ID)) {
			// Not an OAW project yet -- ask user for consent
			final boolean[] faOK = { false };
			Display d;
			if (shell != null) {
				d = shell.getDisplay();
			} else {
				d = Display.getDefault();
			}
			d.syncExec(new Runnable() {

				@Override
				public void run() {
					faOK[0] = MessageDialog
							.openConfirm(
									shell,
									"Converting to oAW project",
									"This will convert project "
											+ project.getName()
											+ " into an openArchitectureWare project. Do you agree?");

				}
			});
			if (!faOK[0]) {
				return false;
			}
		} else {
			// Already an oAW project. Nothing to do...
			return true;
		}

		try {
			monitor.beginTask("", 10);

			// Fix project natures
			int nToAdd = 1;
			boolean fHasJava = true;
			boolean fHasPlugin = true;
			if (!projectDescription.hasNature(JavaCore.NATURE_ID)) {
				fHasJava = false;
				nToAdd++;
			}
			if (!projectDescription.hasNature("org.eclipse.pde.PluginNature")) {
				fHasPlugin = false;
				nToAdd++;
			}
			String[] saNatureIds = new String[projectDescription.getNatureIds().length
					+ nToAdd];
			System.arraycopy(projectDescription.getNatureIds(), 0, saNatureIds,
					nToAdd, projectDescription.getNatureIds().length);
			saNatureIds[0] = OawNature.NATURE_ID;
			int i = 1;
			if (!fHasJava) {
				saNatureIds[i++] = JavaCore.NATURE_ID;
			}
			if (!fHasPlugin) {
				saNatureIds[i++] = "org.eclipse.pde.PluginNature";
			}
			projectDescription.setNatureIds(saNatureIds);

			// Fix project build spec
			if (!fHasPlugin) {
				nToAdd++; // plugin projects actually need two builders
			}
			ICommand[] caBuildSpec = new ICommand[projectDescription
					.getBuildSpec().length
					+ nToAdd];
			System.arraycopy(projectDescription.getBuildSpec(), 0, caBuildSpec,
					nToAdd, projectDescription.getBuildSpec().length);
			caBuildSpec[0] = projectDescription.newCommand();
			caBuildSpec[0].setBuilderName(OawBuilder.getBUILDER_ID());
			i = 1;
			if (!fHasJava) {
				caBuildSpec[i] = projectDescription.newCommand();
				caBuildSpec[i].setBuilderName(JavaCore.BUILDER_ID);
				i++;
			}
			if (!fHasPlugin) {
				caBuildSpec[i] = projectDescription.newCommand();
				caBuildSpec[i]
						.setBuilderName("org.eclipse.pde.ManifestBuilder");
				i++;

				caBuildSpec[i] = projectDescription.newCommand();
				caBuildSpec[i].setBuilderName("org.eclipse.pde.SchemaBuilder");
			}
			projectDescription.setBuildSpec(caBuildSpec);

			// Save the project description
			/* project.open(new SubProgressMonitor(progressMonitor, 1)); */
			project.setDescription(projectDescription, new SubProgressMonitor(
					monitor, 1));

			// Fix up class path and source folders
			IJavaProject javaProject = JavaCore.create(project);
			List<IClasspathEntry> newClassPathEntries = new ArrayList<IClasspathEntry>();

			IFolder srcFolder = project.getFolder("src");
			if (!srcFolder.exists()) {
				srcFolder.create(false, true,
						new SubProgressMonitor(monitor, 1));
				newClassPathEntries.add(JavaCore.newSourceEntry(srcFolder
						.getFullPath()));
			}

			IFolder srcGenFolder = project.getFolder("src-gen");
			if (!srcGenFolder.exists()) {
				srcGenFolder.create(false, true, new SubProgressMonitor(
						monitor, 1));
				newClassPathEntries.add(JavaCore.newSourceEntry(srcGenFolder
						.getFullPath()));
			}

			if (!fHasJava) {
				newClassPathEntries
						.add(JavaCore
								.newContainerEntry(new Path(
										"org.eclipse.jdt.launching.JRE_CONTAINER/org.eclipse.jdt.internal.debug.ui.launcher.StandardVMType/J2SE-1.5")));
			}
			if (!fHasPlugin) {
				newClassPathEntries.add(JavaCore.newContainerEntry(new Path(
						"org.eclipse.pde.core.requiredPlugins")));
			}
			IClasspathEntry[] ceaRawClasspath;
			if (fHasJava) {
				ceaRawClasspath = new IClasspathEntry[javaProject
						.getRawClasspath().length
						+ newClassPathEntries.size()];
				System.arraycopy(javaProject.getRawClasspath(), 0,
						ceaRawClasspath, newClassPathEntries.size(),
						javaProject.getRawClasspath().length);
			} else {
				// Ignore the initial class path set up by JavaProject, as this
				// will cause trouble...
				ceaRawClasspath = new IClasspathEntry[newClassPathEntries
						.size()];

			}
			i = 0;
			for (IClasspathEntry cp : newClassPathEntries) {
				ceaRawClasspath[i++] = cp;
			}
			javaProject.setRawClasspath(ceaRawClasspath,
					new SubProgressMonitor(monitor, 1));

			javaProject.setOutputLocation(new Path("/" + project.getName()
					+ "/bin"), new SubProgressMonitor(monitor, 1));

			IFile fManifest = project.getFile("META-INF/MANIFEST.MF");
			if (!fManifest.exists()) {
				createManifest(project.getName(), monitor, project);
			}
			/*
			 * createBuildProps(progressMonitor, project, srcFolders);
			 */
		} catch (Exception exception) {
			exception.printStackTrace();
			return false;
		} finally {
			monitor.done();
		}

		return true;
	}

	private static void createManifest(String projectName,
			IProgressMonitor monitor, IProject project) throws CoreException {
		final StringBuilder maniContent = new StringBuilder(
				"Manifest-Version: 1.0\n");
		maniContent.append("Bundle-ManifestVersion: 2\n");
		maniContent.append("Bundle-Name: " + projectName + "\n");
		maniContent.append("Bundle-SymbolicName: " + projectName
				+ "; singleton:=true\n");
		maniContent.append("Bundle-Version: 1.0.0\n");
		// maniContent.append("Bundle-Localization: plugin\n");
		maniContent.append("Require-Bundle: ");
		/*
		 * for (final String entry : requiredBundles) { maniContent.append(" " +
		 * entry + ",\n"); }
		 */
		maniContent.append(" org.openarchitectureware.dependencies\n");

		maniContent.append("Bundle-RequiredExecutionEnvironment: J2SE-1.5\r\n");

		final IFolder metaInf = project.getFolder("META-INF");
		metaInf.create(false, true, new SubProgressMonitor(monitor, 1));
		IFile manifest = metaInf.getFile("MANIFEST.MF");
		manifest.create(new StringBufferInputStream(maniContent.toString()),
				true, new SubProgressMonitor(monitor, 1));
	}

	/**
	 * Copy libraries contained in a set of required bundles and make them
	 * direct entries on the classpath of the given project. This is stupid, but
	 * needed so that oAW will find them and be able to use them...
	 * 
	 * @param project
	 * @param monitor
	 * @param saRequiredBundles
	 *            an array of bundle IDs
	 * @throws CoreException 
	 * @throws IOException 
	 * @throws MalformedURLException 
	 */
	public static void addRequiredLibraries(IProject project,
			IProgressMonitor monitor, String[] saRequiredBundles) throws CoreException, MalformedURLException, IOException {
		// Create libs folder in the project
		IFolder fLibsFolder = project.getFolder("/libs");
		if (!fLibsFolder.exists()) {
			fLibsFolder.create(true, true, monitor);
		}

		List<IClasspathEntry> lceClasspath = new ArrayList<IClasspathEntry>();
		for (String sBundle : saRequiredBundles) {
			// Read Manifest file to find additional libraries
			URL urlManifest = FileLocator.toFileURL(new URL("platform:/plugin/"
					+ sBundle + "/META-INF/MANIFEST.MF"));
			Manifest mf = new Manifest(urlManifest.openStream());
			String sBundleCP = mf.getMainAttributes().getValue(
					"Bundle-ClassPath");

			if (sBundleCP != null) {
				// A bundle classpath was defined, analyse it
				for (String sCPElement : sBundleCP.split(",")) {
					sCPElement = sCPElement.trim();
					if (sCPElement.endsWith(".jar")) {
						// A jar file .. copy it
						if (!sCPElement.startsWith("/")) {
							sCPElement = "/" + sCPElement;
						}

						URL urlCPElement = FileLocator.toFileURL(new URL(
								"platform:/plugin/" + sBundle + sCPElement));
						IFile fTarget = fLibsFolder.getFile(sCPElement
								.substring(sCPElement.lastIndexOf("/")));
						if (!fTarget.exists()) {
							fTarget.create(urlCPElement.openStream(), true,
									monitor);
							lceClasspath.add(JavaCore.newLibraryEntry(fTarget
									.getFullPath(), null, null));
						}
					}
				}
			}
		}
		if (!lceClasspath.isEmpty()) {
			// Actually extend project classpath
			IJavaProject jp = JavaCore.create(project);
			IClasspathEntry[] cpOriginal = jp.getRawClasspath();
			IClasspathEntry[] cpNew = new IClasspathEntry[cpOriginal.length
					+ lceClasspath.size()];
			System.arraycopy(cpOriginal, 0, cpNew, lceClasspath.size(),
					cpOriginal.length);
			System.arraycopy(lceClasspath.toArray(), 0, cpNew, 0, lceClasspath
					.size());
			jp.setRawClasspath(cpNew, monitor);
		}
	}
}